@@ -20,11 +20,9 @@
 
 #include <cmath>
 #include <cstdlib>
-#ifndef CLI
+#ifndef HAVE_MPI
 #include <stdexcept>
-#else
-#include <sstream>
-#endif  // CLI
+#endif
 
 #include "somoclu.h"
 
@@ -68,8 +66,7 @@ double get_wall_time(){
 #include <iomanip>
 #endif  // HAVE_R
 
-void my_abort(string err) {
-#ifdef CLI
+void cuda_abort(string err) {
 #ifdef HAVE_MPI
     int rank = 0;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
@@ -77,13 +74,9 @@ void my_abort(string err) {
         cerr << "Error: " << err << endl;
     }
     MPI_Abort(MPI_COMM_WORLD, 1);
-#else
-    cerr << "Error: " << err << endl;
-    exit(1);
-#endif  // HAVE_MPI
 #else
     throw std::runtime_error(err);
-#endif  // CLI
+#endif  // HAVE_MPI
 }
 
 void train(float *data, int data_length, unsigned int nEpoch,
@@ -99,93 +92,41 @@ void train(float *data, int data_length, unsigned int nEpoch,
            float *uMatrix, int uMatrix_size) {
 #ifdef HAVE_R
 #ifndef CUDA
-	if(kernelType == DENSE_GPU){
-		Rprintf("Error: CUDA kernel not compiled \n");
-		return;
-	}
+    if(kernelType == DENSE_GPU){
+        Rprintf("Error: CUDA kernel not compiled \n");
+        return;
+    }
 #endif // CUDA
 #endif // HAVE_R
-    train(0, data, NULL, codebook, globalBmus, uMatrix, nSomX, nSomY,
-          nDimensions, nVectors, nVectors,
-          nEpoch, radius0, radiusN, radiusCooling,
+    som map = {
+      .nSomX = nSomX,
+      .nSomY = nSomY,
+      .nDimensions = nDimensions,
+      .nVectors = nVectors,
+      .mapType = mapType,
+      .gridType = gridType,
+      .get_distance = EuclideanDistance(nDimensions),
+      .uMatrix = uMatrix,
+      .codebook = codebook,
+      .bmus = globalBmus};
+    
+    train(0, data, NULL, map, nVectors, nEpoch, radius0, radiusN, radiusCooling,
           scale0, scaleN, scaleCooling,
-          kernelType, mapType,
-          gridType, compact_support, gaussian, std_coeff, verbose
-#ifdef CLI
-          , "", 0);
-#else
-         );
-#endif
-    calculateUMatrix(uMatrix, codebook, nSomX, nSomY, nDimensions, mapType,
-                     gridType);
+          kernelType, compact_support, gaussian, std_coeff, verbose);
+    calculateUMatrix(map);
 }
 
-void julia_train(float *data, int data_length, unsigned int nEpoch,
-           unsigned int nSomX, unsigned int nSomY,
-           unsigned int nDimensions, unsigned int nVectors,
-           float radius0, float radiusN, unsigned int _radiusCooling,
-           float scale0, float scaleN, unsigned int _scaleCooling,
-           unsigned int kernelType, unsigned int _mapType,
-           unsigned int _gridType, bool compact_support, bool gaussian,
-           float std_coeff, unsigned int verbose,
-           float *codebook, int codebook_size,
-           int *globalBmus, int globalBmus_size,
-           float *uMatrix, int uMatrix_size) {
-    string radiusCooling;
-    string scaleCooling;
-    string mapType;
-    string gridType;
-    if (_radiusCooling == 0) {
-        radiusCooling = "linear";
-    } else {
-        radiusCooling = "exponential";
-    }
-    if (_scaleCooling == 0) {
-        scaleCooling = "linear";
-    } else {
-        scaleCooling = "exponential";
-    }
-    if (_mapType == 0) {
-        mapType = "planar";
-    } else {
-        mapType = "toroid";
-    }
-    if (_gridType == 0) {
-        gridType = "square";
-    } else {
-        gridType = "hexagonal";
-    }
-    train(0, data, NULL, codebook, globalBmus, uMatrix, nSomX, nSomY,
-          nDimensions, nVectors, nVectors,
-          nEpoch, radius0, radiusN, radiusCooling,
-          scale0, scaleN, scaleCooling,
-          kernelType, mapType,
-          gridType, compact_support, gaussian, std_coeff, verbose
-#ifdef CLI
-          , "", 0);
-#else
-         );
-#endif
-    calculateUMatrix(uMatrix, codebook, nSomX, nSomY, nDimensions, mapType,
-                     gridType);
-}
 
 void train(int itask, float *data, svm_node **sparseData,
-           float *codebook, int *globalBmus, float *uMatrix,
-           unsigned int nSomX, unsigned int nSomY,
-           unsigned int nDimensions, unsigned int nVectors,
-           unsigned int nVectorsPerRank, unsigned int nEpoch,
-           float radius0, float radiusN, string radiusCooling,
-           float scale0, float scaleN, string scaleCooling,
-           unsigned int kernelType, string mapType,
-           string gridType, bool compact_support, bool gaussian,
-           float std_coeff, unsigned int verbose
-#ifdef CLI
-           , string outPrefix, unsigned int snapshots)
-#else
-          )
-#endif
+           som map, unsigned int nVectorsPerRank, unsigned int nEpoch,
+           float radius0, float radiusN,
+           string radiusCooling,
+           float scale0, float scaleN,
+           string scaleCooling,
+           unsigned int kernelType, bool compact_support, bool gaussian,
+           float std_coeff, unsigned int verbose, Snapshot *snapshot)
 {
+    float * X2 = NULL;
     int nProcs = 1;
 #ifdef HAVE_MPI
     MPI_Comm_size(MPI_COMM_WORLD, &nProcs);
@@ -193,21 +134,39 @@ void train(int itask, float *data, svm_node **sparseData,
 #ifdef CUDA
     if (kernelType == DENSE_GPU) {
         setDevice(itask, nProcs);
-        initializeGpu(data, nVectorsPerRank, nDimensions, nSomX, nSomY);
+        initializeGpu(data, nVectorsPerRank, map);
     }
 #endif
-    // (Re-)Initialize codebook with random values only if requested through
-    // the passed codebook -- meaning that the user did not have an initial
-    // codebook
 
-    if (codebook[0] == 1000 && codebook[1] == 2000) {
-        initializeCodebook(get_wall_time(), codebook, nSomX, nSomY, nDimensions);
+    if (kernelType == SPARSE_CPU) {
+        // Pre-compute the squared norm of all the vectors
+
+        X2 = new float[nVectorsPerRank];
+
+        #pragma omp parallel for
+        for (omp_iter_t i=0; i<nVectorsPerRank; ++i) {
+            if (itask * nVectorsPerRank + i < map.nVectors) {
+                float acc=0.f;
+                for (unsigned int j=0; sparseData[i][j].index!=-1; ++j) {
+                    acc += sparseData[i][j].value * sparseData[i][j].value;
+                }
+                X2[i] = acc;
+            }
+        }
+    }
+
+    // (Re-)Initialize map.codebook with random values only if requested through
+    // the passed map.codebook -- meaning that the user did not have an initial
+    // map.codebook
+
+    if (map.codebook[0] == 1000 && map.codebook[1] == 2000) {
+        initializeCodebook(get_wall_time(), map);
     }
     ///
     /// Parameters for SOM
     ///
     if (radius0 == 0) {
-        unsigned int minDim = min(nSomX, nSomY);
+        unsigned int minDim = min(map.nSomX, map.nSomY);
         radius0 = minDim / 2.0f;              /// init radius for updating neighbors
     }
     if (radiusN == 0) {
@@ -224,26 +183,15 @@ void train(int itask, float *data, svm_node **sparseData,
     while ( currentEpoch < nEpoch ) {
 
         double epoch_time = get_wall_time();
-        trainOneEpoch(itask, data, sparseData, codebook, globalBmus,
-                      nEpoch, currentEpoch,
-                      nSomX, nSomY, nDimensions, nVectors, nVectorsPerRank,
+        trainOneEpoch(itask, data, sparseData, X2, map,
+                      nEpoch, currentEpoch, nVectorsPerRank,
                       radius0, radiusN, radiusCooling,
-                      scale0, scaleN, scaleCooling, kernelType, mapType,
-                      gridType, compact_support, gaussian, std_coeff);
-#ifdef CLI
-        if (snapshots > 0 && itask == 0) {
-            calculateUMatrix(uMatrix, codebook, nSomX, nSomY, nDimensions,
-                             mapType, gridType);
-            stringstream sstm;
-            sstm << outPrefix << "." << currentEpoch + 1;
-            saveUMatrix(sstm.str() + string(".umx"), uMatrix, nSomX, nSomY);
-            if (snapshots == 2) {
-                saveBmus(sstm.str() + string(".bm"), globalBmus, nSomX, nSomY, nVectors);
-                saveCodebook(sstm.str() + string(".wts"), codebook, nSomX, nSomY, nDimensions);
-            }
-        }
-#endif
+                      scale0, scaleN, scaleCooling, kernelType, compact_support,
+                      gaussian, std_coeff);
         ++currentEpoch;
+        if (snapshot != NULL && itask == 0) {
+            snapshot->write(currentEpoch, map);
+        }
 #ifndef HAVE_R
         if (itask == 0 && verbose > 0) {
             epoch_time = get_wall_time() - epoch_time;
@@ -268,17 +216,20 @@ void train(int itask, float *data, svm_node **sparseData,
         cout << endl;
     }
 #endif
-    trainOneEpoch(itask, data, sparseData, codebook, globalBmus,
-                  nEpoch, currentEpoch,
-                  nSomX, nSomY, nDimensions, nVectors, nVectorsPerRank,
+    trainOneEpoch(itask, data, sparseData, X2, map,
+                  nEpoch, currentEpoch,nVectorsPerRank,
                   radius0, radiusN, radiusCooling,
-                  scale0, scaleN, scaleCooling, kernelType, mapType,
-                  gridType, compact_support, gaussian, std_coeff, true);
+                  scale0, scaleN, scaleCooling, kernelType, compact_support,
+                  gaussian, std_coeff, true);
 #ifdef CUDA
     if (kernelType == DENSE_GPU) {
         freeGpu();
     }
 #endif
+
+    if (kernelType == SPARSE_CPU) {
+        delete [] X2;
+    }
 }
 
 float linearCooling(float start, float end, float nEpoch, float epoch) {
@@ -299,16 +250,15 @@ float exponentialCooling(float start, float end, float nEpoch, float epoch) {
 
 
 
-/** Initialize SOM codebook with random values
+/** Initialize SOM map.codebook with random values
  * @param seed - random seed
- * @param codebook - the codebook to fill in
- * @param nSomX - dimensions of SOM map in the currentEpoch direction
- * @param nSomY - dimensions of SOM map in the y direction
- * @param nDimensions - dimensions of a data instance
+ * @param map.codebook - the map.codebook to fill in
+ * @param map.nSomX - dimensions of SOM map in the currentEpoch direction
+ * @param map.nSomY - dimensions of SOM map in the y direction
+ * @param map.nDimensions - dimensions of a data instance
  */
 
-void initializeCodebook(unsigned int seed, float *codebook, unsigned int nSomX,
-                        unsigned int nSomY, unsigned int nDimensions) {
+void initializeCodebook(unsigned int seed, som map) {
     ///
     /// Fill initial random weights
     ///
@@ -318,20 +268,16 @@ void initializeCodebook(unsigned int seed, float *codebook, unsigned int nSomX,
     srand(seed);
 #endif
     #pragma omp parallel for
-#ifdef _WIN32
-    for (int som_y = 0; som_y < nSomY; som_y++) {
-#else
-	for (unsigned int som_y = 0; som_y < nSomY; som_y++) {
-#endif
-		for (unsigned int som_x = 0; som_x < nSomX; som_x++) {
-			for (unsigned int d = 0; d < nDimensions; d++) {
+    for (omp_iter_t som_y = 0; som_y < map.nSomY; som_y++) {
+        for (unsigned int som_x = 0; som_x < map.nSomX; som_x++) {
+            for (unsigned int d = 0; d < map.nDimensions; d++) {
 #ifdef HAVE_R
                 int w = 0xFFF & (int) (RAND_MAX * R::runif(0,1));
 #else
                 int w = 0xFFF & rand();
 #endif
                 w -= 0x800;
-                codebook[som_y * nSomX * nDimensions + som_x * nDimensions + d] = (float)w / 4096.0f;
+                map.codebook[som_y * map.nSomX * map.nDimensions + som_x * map.nDimensions + d] = (float)w / 4096.0f;
             }
         }
     }
@@ -340,34 +286,30 @@ void initializeCodebook(unsigned int seed, float *codebook, unsigned int nSomX,
 #endif
 }
 
-void trainOneEpoch(int itask, float *data, svm_node **sparseData,
-                   float *codebook, int *globalBmus,
-                   unsigned int nEpoch, unsigned int currentEpoch,
-                   unsigned int nSomX, unsigned int nSomY,
-                   unsigned int nDimensions, unsigned int nVectors,
+void trainOneEpoch(int itask, float *data, svm_node **sparseData, float *X2,
+                   som map, unsigned int nEpoch, unsigned int currentEpoch,
                    unsigned int nVectorsPerRank,
                    float radius0, float radiusN,
                    string radiusCooling,
                    float scale0, float scaleN,
                    string scaleCooling,
-                   unsigned int kernelType, string mapType,
-                   string gridType, bool compact_support, bool gaussian,
+                   unsigned int kernelType, bool compact_support, bool gaussian,
                    float std_coeff, bool only_bmus) {
 
     float N = (float)nEpoch;
-    float *numerator;
-    float *denominator;
+    float *numerator = NULL;
+    float *denominator = NULL;
     float scale = scale0;
     float radius = radius0;
     if (itask == 0 && !only_bmus) {
 #ifdef HAVE_MPI
-        numerator = new float[nSomY * nSomX * nDimensions];
-        denominator = new float[nSomY * nSomX];
-        for (unsigned int som_y = 0; som_y < nSomY; som_y++) {
-            for (unsigned int som_x = 0; som_x < nSomX; som_x++) {
-                denominator[som_y * nSomX + som_x] = 0.0;
-                for (unsigned int d = 0; d < nDimensions; d++) {
-                    numerator[som_y * nSomX * nDimensions + som_x * nDimensions + d] = 0.0;
+        numerator = new float[map.nSomY * map.nSomX * map.nDimensions];
+        denominator = new float[map.nSomY * map.nSomX];
+        for (unsigned int som_y = 0; som_y < map.nSomY; som_y++) {
+            for (unsigned int som_x = 0; som_x < map.nSomX; som_x++) {
+                denominator[som_y * map.nSomX + som_x] = 0.0;
+                for (unsigned int d = 0; d < map.nDimensions; d++) {
+                    numerator[som_y * map.nSomX * map.nDimensions + som_x * map.nDimensions + d] = 0.0;
                 }
             }
         }
@@ -390,7 +332,7 @@ void trainOneEpoch(int itask, float *data, svm_node **sparseData,
     if (!only_bmus) {
         MPI_Bcast(&radius, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
         MPI_Bcast(&scale, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
-        MPI_Bcast(codebook, nSomY * nSomX * nDimensions, MPI_FLOAT,
+        MPI_Bcast(map.codebook, map.nSomY * map.nSomX * map.nDimensions, MPI_FLOAT,
                   0, MPI_COMM_WORLD);
     }
 #endif
@@ -402,48 +344,39 @@ void trainOneEpoch(int itask, float *data, svm_node **sparseData,
     default:
     case DENSE_CPU:
         trainOneEpochDenseCPU(itask, data, numerator, denominator,
-                              codebook, nSomX, nSomY, nDimensions,
-                              nVectors, nVectorsPerRank, radius, scale,
-                              mapType, gridType, compact_support, gaussian,
-                              globalBmus, only_bmus, std_coeff);
+                              map, nVectorsPerRank, radius, scale,
+                              compact_support, gaussian, only_bmus, std_coeff);
         break;
     case DENSE_GPU:
 #ifdef CUDA
         trainOneEpochDenseGPU(itask, data, numerator, denominator,
-                              codebook, nSomX, nSomY, nDimensions,
-                              nVectors, nVectorsPerRank, radius, scale,
-                              mapType, gridType, compact_support, gaussian, globalBmus, only_bmus, std_coeff);
+                              map, nVectorsPerRank, radius, scale,
+                              compact_support, gaussian, only_bmus, std_coeff);
 #else
-        my_abort("Compiled without CUDA!");
+        cuda_abort("Compiled without CUDA!");
 #endif
         break;
     case SPARSE_CPU:
-        trainOneEpochSparseCPU(itask, sparseData, numerator, denominator,
-                               codebook, nSomX, nSomY, nDimensions,
-                               nVectors, nVectorsPerRank, radius, scale,
-                               mapType, gridType, compact_support, gaussian,
-                               globalBmus, only_bmus, std_coeff);
+        trainOneEpochSparseCPU(itask, sparseData, X2, numerator, denominator,
+                               map, nVectorsPerRank, radius, scale,
+                               compact_support, gaussian, only_bmus, std_coeff);
         break;
     }
 
-    /// 3. Update codebook using numerator and denominator
+    /// 3. Update map.codebook using numerator and denominator
 #ifdef HAVE_MPI
     if (!only_bmus) {
       MPI_Barrier(MPI_COMM_WORLD);
       if (itask == 0 && !only_bmus) {
           #pragma omp parallel for
-#ifdef _WIN32
-          for (int som_y = 0; som_y < nSomY; som_y++) {
-#else
-          for (unsigned int som_y = 0; som_y < nSomY; som_y++) {
-#endif
-              for (unsigned int som_x = 0; som_x < nSomX; som_x++) {
-                  float denom = denominator[som_y * nSomX + som_x];
+          for (omp_iter_t som_y = 0; som_y < map.nSomY; som_y++) {
+              for (unsigned int som_x = 0; som_x < map.nSomX; som_x++) {
+                  float denom = denominator[som_y * map.nSomX + som_x];
                   if (denom != 0) {
-                    for (unsigned int d = 0; d < nDimensions; d++) {
-                        float newWeight = numerator[som_y * nSomX * nDimensions
-                                                    + som_x * nDimensions + d] / denom;
-                        codebook[som_y * nSomX * nDimensions + som_x * nDimensions + d] = newWeight;
+                    for (unsigned int d = 0; d < map.nDimensions; d++) {
+                        float newWeight = numerator[som_y * map.nSomX * map.nDimensions
+                                                    + som_x * map.nDimensions + d] / denom;
+                        map.codebook[som_y * map.nSomX * map.nDimensions + som_x * map.nDimensions + d] = newWeight;
                     }
                   }
               }
